# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
import logging
from django.db import connections


def index_exists(index, database='default'):
    """Execute raw sql
    """
    cursor = connections[database].cursor()
    qry = "SELECT COUNT(indexname) FROM pg_indexes WHERE indexname = %s"
    cursor.execute(qry, [index['name']])
    row = cursor.fetchone()
    cursor.close()
    return row[0] == 1


def execute_raw(sql, database='default', parms=None):
    """
    Execute a raw SQL command

    sql (string) : SQL command
    database (string): the database name configured in settings
    """
    try:
        cursor = connections[database].cursor()
        if parms is not None:
            cursor.execute(sql)
        else:
            cursor.execute(sql, parms)            
        cursor.close()
        return 0
    except Exception, e:
        logging.error('Cant execute %s -- Exception raised %s' % (sql, e))
        return 1


def drop_index(index, database='default'):
    """
    Check if index exists and drop it

    index (dict) : index description
    """
    if 'database' in index:
        database = index['database']

    if index_exists(index, database):
        logging.info("Will drop %s" % index['name'])

        res = execute_raw(index['cmd'], database)

        logging.info("%s dropped" % index['name'])
    else:
        res = 1
        logging.info("%s doesn't exists" % index['name'])
    return res


def create_index(index, database='default'):
    """
    Create an index

    index (dict) : index description
       {"name": "foo",
        "database": "default",
        "cmd": "CREATE INDEX foo_idx ON table (column)"
       }
    """
    if 'database' in index:
        database = index['database']

    if index_exists(index, database):
        logging.info("%s still exists" % index['name'])
        res = 1
    else:
        logging.info("Will create %s" % index['name'])
        res = execute_raw(index['cmd'], database)
        logging.info("%s created" % index['name'])
    return res


def create_extensions(extensions, database='default'):
    """
    Create all extensions
    """
    if 'database' in index:
        database = index['database']

    for extension in extensions:
        cmd = "CREATE EXTENSION IF NOT EXISTS %s"
        logging.info("Will create extension %s on database %s" % (extension, database))
        res = execute_raw(cmd,
                          database=database,
                          parms=[extension])
        logging.info("%s created" % extension)
    return res



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
from django.core.management.base import BaseCommand
from ... import util


class Command(BaseCommand):
    help = 'List index'

    def handle(self, *args, **options):
        """List indexes command on stdout
        Read the table book without TextField
        """
        output = util.command_list()
        if output:
            self.stdout.write(output)



"""
django-json-dbindex tests
"""
import os
import util
from django.test import TestCase


class SimpleTest(TestCase):

    def test_sql_simple(self):
        """
        Return of sql_simple
        """
        idx = {'foo': 'bar'}
        res = "FOOBAR bar"
        self.assertEqual(util.sql_simple(idx, 'foo', 'FOOBAR'), res)

    def test_sql_simple_nokey(self):
        """
        The looking key is absent
        """
        idx = {'key': 'bar'}
        res = util.sql_simple(idx, 'foo', 'FOOBAR')
        attend = ""
        self.assertEqual(res, attend)

    def test_sql_columns(self):
        """
        Build the column part of index
        """
        idx = {'foo': 'bar',
               'columns': ['foobar', 'id']}
        res = "(foobar,id)"
        self.assertEqual(util.sql_columns(idx), res)

    def test_sql_using(self):
        """
        Build the column part of index
        """
        idx = {'using': 'btree'}

        res = "USING btree"
        self.assertEqual(util.sql_using(idx), res)

    def test_sql_predicat(self):
        """
        Build the column part of index
        """
        idx = {'using': 'btree',
               'predicat': 'foo > 3'}

        res = "WHERE foo > 3"
        self.assertEqual(util.sql_predicat(idx), res)

    def test_sql_tablespace(self):
        """
        Build the column part of index
        """
        idx = {'tablespace': 'ssd1'}

        res = "TABLESPACE ssd1"
        self.assertEqual(util.sql_tablespace(idx), res)

    def test_sql_unique(self):
        """
        Build the column part of index
        """
        idx = {'unique': True}

        res = "UNIQUE"
        self.assertEqual(util.sql_unique(idx), res)

    def test_sql_create_from_json(self):
        """
        Build the column part of index
        """
        idx = {'name': 'compo1',
               'table': 'editors',
               'columns': ['id', 'name'],
               'tablespace': 'ssd1'}

        res = " ".join(["CREATE INDEX CONCURRENTLY compo1",
                        "ON editors (id,name)",
                        "TABLESPACE ssd1"])
        self.assertEqual(util.sql_create_from_json(idx), res)

    def test_sql_create_from_json_opclass(self):
        """
        Build the column part of index with an operator class
        """
        idx = {'name': 'compo1',
               'table': 'editors',
               'columns': [{'name': 'gist_trgm_ops'}],
               'tablespace': 'ssd1',
               'using': 'GIST'}

        res = " ".join(["CREATE INDEX CONCURRENTLY compo1",
                        "ON editors USING GIST (name gist_trgm_ops)",
                        "TABLESPACE ssd1"])
        self.assertEqual(util.sql_create_from_json(idx), res)

    def test_sql_drop_from_json(self):
        """
        Build the column part of index
        """
        idx = {'name': 'compo1',
               'table': 'editors',
               'columns': ['id', 'name'],
               'tablespace': 'ssd1'}

        cmd = util.sql_drop_from_json(idx)

        res = 'DROP INDEX compo1'

        self.assertEqual(cmd, res)

    def test_sql_create_from_json_full(self):
        """
        Build the column part of index
        """
        idx = {'name': 'compo1',
               'table': 'editors',
               'unique': True,
               'columns': ['id', 'name'],
               'tablespace': 'ssd1'}

        res = " ".join(["CREATE UNIQUE INDEX CONCURRENTLY compo1",
                        "ON editors (id,name)",
                        "TABLESPACE ssd1"])
        self.assertEqual(util.sql_create_from_json(idx), res)

    def test_list_index_create(self):
        """
        List index to be created
        """
        fpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'samples')
        res = util.list_indexes_create(fpath)
        self.assertEqual(len(res), 1)

    def test_list_index_drop(self):
        """
        List index to be droped
        """
        fpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'samples')
        res = util.list_indexes_drop(fpath)
        self.assertEqual(len(res), 2)

    def test_list_indexes(self):
        """
        List all indexes
        """
        fpath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'samples')
        res = util.list_indexes(fpath)
        self.assertEqual(len(res), 3)


    def test_list_extentions(self):
        """
        List all extensions
        """
        idx = [{'name': 'compo1',
                'table': 'editors',
                'columns': ['id', 'name'],
                'tablespace': 'ssd1',
                'extension': 'unaccent'},
               {'name': 'compo2',
                'table': 'editors',
                'columns': ['id', 'name'],
                'tablespace': 'ssd1',
                'extension': 'unaccent'},
               {'name': 'compo3',
                'table': 'editors',
                'columns': ['id', 'name'],
                'tablespace': 'ssd1',
                'extension': 'pg_trgm'},
               {'name': 'compo4',
                'table': 'editors',
                'columns': ['id', 'name'],
                'tablespace': 'ssd1'}]
        
        extensions = util.list_extensions(idx)

        self.assertEqual(len(extensions), 2)
        self.assertTrue('unaccent' in extensions)
        self.assertTrue('pg_trgm' in extensions)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of django-json-dbindex nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.
#
from django.utils.importlib import import_module
from django.conf import settings
import json
import sys
import re
import logging
import pgcommands
from os import path

FILENAME_CREATE = 'dbindex_create.json'
FILENAME_DROP = 'dbindex_drop.json'


def command_check():
    """
    Check indexes
    """
    output = []
    for fpath in get_app_paths():
        for index in list_indexes_create(fpath):
            if 'database' in index:
                database = index['database']
            else:
                database = 'default'
            if pgcommands.index_exists(index, database):
                msg = "OK %s is present on %s in database %s"
                output.append(msg % (index['name'], index['table'], database))
            else:
                msg = "KO %s is missing (must be present) in database %s"
                output.append(msg % (index['name'], database))

        for index in list_indexes_drop(fpath):
            if pgcommands.index_exists(index):
                output.append("KO %s is present (must be dropped)"
                              % (index['name']))
            else:
                output.append("OK %s is missing" % (index['name']))
    return "\n".join(output)


def command_list():
    """
    List all indexes
    """
    output = []
    for fpath in get_app_paths():
        indexes = list_indexes(fpath)
        if len(indexes):
            output.append("-- Found %d index in %s\n" % (len(indexes), fpath))
            for index in indexes:
                output.append("%s\n" % (index['cmd']))
    return "\n".join(output)


def command_drop():
    """
    The drop command
    """
    # loop on apps
    for fpath in get_app_paths():
        # loop on indexes
        for index in list_indexes_drop(fpath):
            pgcommands.drop_index(index)


def command_create():
    """
    Create all indexes
    """
    for fpath in get_app_paths():
        indexes = list_indexes_create(fpath)
        extensions = list_extensions(indexes)
        for extension in extensions:
            pgcommands.create_extension(index)

        for index in indexes:
            pgcommands.create_index(index)


def get_app_paths():
    """
    Return all paths defined in settings
    """
    fpaths = []
    for app in settings.INSTALLED_APPS:
        try:
            import_module(app)
            fpaths.append(sys.modules[app].__path__[0])
        except AttributeError:
            msg = "Can't load module %s"
            logging.error(msg % app)
            continue
    return fpaths


def list_indexes(fpath):
    """
    List ALL indexes
    """
    return list_indexes_drop(fpath) + list_indexes_create(fpath)


def list_indexes_create(fpath):
    """
    Read indexes

    return (array)
    """
    indexes = []

    pgpath = path.join(fpath, FILENAME_CREATE)
    logging.debug(pgpath)
    if path.isfile(pgpath):
        with open(pgpath) as json_data:
            indexes = json.load(json_data)
            for index in indexes:
                index['cmd'] = sql_create_from_json(index)
            json_data.close()
    return indexes


def list_indexes_drop(fpath):
    """
    Read indexes
    """
    indexes = []

    pgpath = path.join(fpath, FILENAME_DROP)
    if path.isfile(pgpath):
        with open(pgpath) as json_data:
            indexes = json.load(json_data)
            for index in indexes:
                index['cmd'] = sql_drop_from_json(index)
            json_data.close()
    return indexes


def sql_create_from_json(index):
    """
    Read indexes
    """
    cmd = " ".join(["CREATE",
                    sql_unique(index),
                    "INDEX",
                    sql_concurrently(index),
                    index['name'],
                    "ON",
                    index['table'],
                    sql_using(index),
                    sql_columns(index),
                    sql_tablespace(index),
                    sql_predicat(index)
                    ])
    strc = re.sub('\s+', ' ', cmd)
    return re.sub('\s$', '', strc)


def sql_drop_from_json(index):
    """
    Read indexes
    """
    cmd = " ".join(["DROP INDEX",
                    index['name']])
    return cmd


def sql_unique(index):
    """
    Is the index unique or not

    return : string
    """
    res = ""
    if "unique" in index:
        if index["unique"]:
            res = "UNIQUE"
    return res


def sql_concurrently(index):
    """
    Is the index concurrently or not

    return : string
    """
    res = "CONCURRENTLY"
    if "concurrently" in index:
        if index["concurrently"] is False:
            res = ""
    return res


def sql_columns(index):
    """
    Is the index using or not

    return : string
    """
    columns = []

    for column in index['columns']:
        if type(column) is dict:
            columns.append("%s %s" % (column.keys()[0],
                                      column.values()[0]))
        elif type(column) is str:
            columns.append(column)

    return "(%s)" % (",".join(columns))


def sql_using(index):
    """
    Is the index using or not

    return : string
    """
    return sql_simple(index, "using", "USING")


def sql_predicat(index):
    """
    Is the index predicat or not

    return : string
    """
    return sql_simple(index, "predicat", "WHERE")


def sql_tablespace(index):
    """
    create the index on a special tablespace

    return : string
    """
    return sql_simple(index, "tablespace", "TABLESPACE")


def sql_simple(index, key, prefix):
    """

    return : string
    """
    res = ""
    if key in index:
        if index[key] is not None:
            res = "%s %s" % (prefix, index[key])

    return res


def list_extensions(indexes):
    """
    List extensions required in indexes

    Return a list with unique extension present in all indexes

    indexes (array)

    Return : array
    """
    extensions = []
    for index in indexes:
        ext = index.get('extension')
        if ext is not None:
            extensions.append(ext)

    return set(extensions)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014 Rodolphe Quiédeville <rodolphe@quiedeville.org>
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from django.core.management.base import BaseCommand
from ... import util


class Command(BaseCommand):
    """
    DROP all indexes defines in dbindex_drop.json file in all apps
    """

    def handle(self, *args, **options):
        """
        Create all indexes
        """
        util.command_drop()
